import ujson

from edge.admin.ui_types import *
from edge.path_helpers import *
from os.path import join as join_path

import edge.pipeline_helpers as helpers


def ra_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :type avaliable_ui_modules: set of UiIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    if ui_entry_point in [Pipeline.Speed.Ui.LOCATION_0]:
        used_ui_modules.update([Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.VIDEOSOURCE_0])
        used_edge_modules.add(Pipeline.Speed.Edge.RADAR_0)
        creator_ctor = RaCreator.create

    elif ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.VIDEOSOURCE_0])
        creator_ctor = RaCreator.create

    return creator_ctor


class RaCreator(BaseCreator):
    def __init__(self, id_main_setupstore):
        """
        :type id_main_setupstore: str, dict, str
        """
        self._id_main_setupstore = id_main_setupstore
        BaseCreator.__init__(self)

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: RaCreator
        """
        if ui_entry_point in [Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.PIPELINE]:
            id_main_setupstore = RaCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map)
            creator = RaCreator(id_main_setupstore)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict | None
        :rtype: str, dict, str
        """
        from _edge_radar_location import RaLocationEditor

        ui_location_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.LOCATION_0]
        ui_videosource_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.VIDEOSOURCE_0]
        edge_identity = Pipeline.Speed.Edge.RADAR_0

        edge_gid = edge_identity.gid
        spool_dir = get_spool_dir(ui_pid)

        edge_group = helpers.get_edge_group(ui_pid)
        edge_id = helpers.get_edge_id(edge_group, edge_identity)

        # send sms to module radar with new config
        if ui_entry_point == Pipeline.Speed.Ui.LOCATION_0:
            radar_lock_dirpath = '/etc/edge/radar.%s/lock' % ui_pid
            editor = RaLocationEditor.deserialize(ui_location_jnode, ui_videosource_jnode, radar_lock_dirpath)
            editor.serialize(ui_pid)

        # create main.conf
        main_conf_jnode = {
            "api-server": {
                "endpoint": join_path(spool_dir, "ra%s-ra%s-uds" % (edge_gid, edge_gid))
            },
            "device": {
                "vendor": "smartmicro",
                "endpoint": ui_location_jnode['ro'].get('uri', 'ummr0a:///dev/ttyS0'),
                "runtime": "/etc/edge/radar.%s/runtime/" % edge_gid
            },
            "strobe-react": {
                "endpoint": join_path(spool_dir, "mg%s-ra%s-pipe" % (edge_gid, edge_gid))
            },
            "scene-event": {
                "endpoints": [
                    join_path(spool_dir, "ra%s-tr%s-scene" % (edge_gid, edge_gid))
                ]
            }
        }

        # create setupstore.conf
        setupstore_conf_text = get_setupstore_conf_text(edge_id)

        # create haltaction.conf
        def_halt_list = [
            "0:python /opt/edge/bin/psuctl run led-radar color G blink 0HZ",
            "1:python /opt/edge/bin/psuctl run led-radar color R blink 1HZ; python /opt/edge/bin/psuctl run radar-reboot",
            "2:python /opt/edge/bin/psuctl run led-radar color R blink 0HZ"
        ]
        halt_list = ui_location_jnode['ro'].get('haltaction', def_halt_list)
        if halt_list:
            haltaction_conf = '\n'.join(halt_list)

        else:
            haltaction_conf = ''

        return edge_id, main_conf_jnode, setupstore_conf_text, haltaction_conf

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text, haltaction_conf_text = self._id_main_setupstore

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_RADAR, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_filepath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_filepath, 'w') as json_file:
            ujson.dump(main_conf_jnode, json_file, indent=2)

        # create setupstore.conf
        setupstore_conf_filepath = join_path(edge_jnode_dir, 'setupstore.conf')
        with open(setupstore_conf_filepath, 'w') as text_file:
            text_file.write(setupstore_conf_text)

        halt_action_path = join_path(edge_jnode_dir, 'haltaction.conf')
        with open(halt_action_path, 'w') as text_file:
            text_file.write(haltaction_conf_text)


class RaDeleter:
    def __init__(self, edge_ra_dirpath):
        """
        :type edge_ra_dirpath: str
        """
        self._edge_ra_dirpath = edge_ra_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: RaDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.RADAR_0)
            edge_type = Pipeline.Speed.Edge.RADAR_0.type
            deleter = RaDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            return None

        elif ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            return None

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_ra_dirpath)
