import ujson
import os

from edge.admin.ui_types import BaseCreator
from edge.path_helpers import *


import edge.pipeline_helpers as helpers


def or_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :type avaliable_ui_modules: set of UiIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    ui_modules_set = set()
    if ui_entry_point.type in [UI_TYPE_LANE_ENFORCEMENT, UI_TYPE_HTTP_STREAMING,
                               UI_TYPE_RTSP_STREAMING, UI_TYPE_SPOT_SPEED,
                               UI_TYPE_PIPELINE, UI_TYPE_GEOLOCATION,
                               UI_TYPE_EVENTSTORE]:
        # define ui_jnodes_set
        for ui_module in avaliable_ui_modules:
            if ui_module.type in [UI_TYPE_LANE_ENFORCEMENT, UI_TYPE_HTTP_STREAMING,
                                  UI_TYPE_RTSP_STREAMING, UI_TYPE_SPOT_SPEED,
                                  UI_TYPE_GEOLOCATION]:
                ui_modules_set.add(ui_module)

        used_ui_modules.update(ui_modules_set)

        # define used edge modules
        if ui_entry_point in [Pipeline.Speed.Ui.PIPELINE, Pipeline.Mv.Ui.PIPELINE,
                              Pipeline.Rtsp.Ui.PIPELINE]:
            pass

        elif ui_entry_point in [Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0, Pipeline.Rtsp.Ui.GEOLOCATION_0,
                                Pipeline.Rtsp.Ui.EVENTSTORE_0]:
            used_edge_modules.add(EdgeIdentity(EDGE_TYPE_RENDERER, '0'))

        elif ui_entry_point in [Pipeline.Speed.Ui.LANE_ENFORCEMENT_0, Pipeline.Mv.Ui.LANE_ENFORCEMENT_0,
                                Pipeline.Speed.Ui.EVENTSTORE_0, Pipeline.Mv.Ui.EVENTSTORE_0,
                                Pipeline.Speed.Ui.GEOLOCATION_0, Pipeline.Mv.Ui.GEOLOCATION_0]:
            used_edge_modules.update([EdgeIdentity(EDGE_TYPE_RENDERER, '0'), EdgeIdentity(EDGE_TYPE_RENDERER, '1')])

        elif ui_entry_point.type == UI_TYPE_HTTP_STREAMING:
            used_edge_modules.add(EdgeIdentity(EDGE_TYPE_RENDERER, '0'))

        elif ui_entry_point.type == UI_TYPE_RTSP_STREAMING:
            used_edge_modules.add(EdgeIdentity(EDGE_TYPE_RENDERER, '1'))

        elif ui_entry_point.type in [UI_TYPE_SPOT_SPEED]:
            used_edge_modules.update([EdgeIdentity(EDGE_TYPE_RENDERER, '0'), EdgeIdentity(EDGE_TYPE_RENDERER, '1')])

        creator_ctor = OrCreator.create
    else:
        return None

    return lambda arg0, arg1, arg2: creator_ctor(arg0, arg1, arg2, ui_modules_set)


class OrCreator(BaseCreator):
    def __init__(self, id_main_setupstore_list):
        """
        :type id_main_setupstore_list: list of (str, dict, str)
        """
        BaseCreator.__init__(self)

        self._id_main_setupstore_list = id_main_setupstore_list

    @staticmethod
    def _create_lpr_rtsp(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: OrCreator | None
        """

        if ui_entry_point in [Pipeline.Rtsp.Ui.PIPELINE, Pipeline.Rtsp.Ui.HTTP_STREAMING_0,
                              Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0, Pipeline.Rtsp.Ui.GEOLOCATION_0,
                              Pipeline.Rtsp.Ui.EVENTSTORE_0]:
            spool_dir = get_spool_dir(ui_pid)

            or0_ws_scene_dir = os.path.join(
                spool_dir, 'or%s-ws%s-scene' % (Pipeline.Rtsp.Edge.RENDERER_0.gid, Pipeline.Rtsp.Edge.WEBSERVER_0.gid))

            tr0_or0_scene_dir = os.path.join(
                spool_dir, 'tr%s-or%s-scene' % (Pipeline.Rtsp.Edge.TRACKER_0.gid, Pipeline.Rtsp.Edge.RENDERER_0.gid))

            or0_args = OrCreator._serialize(
                ui_identity_to_jnode_map[Pipeline.Rtsp.Ui.HTTP_STREAMING_0], None,
                ui_identity_to_jnode_map[Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0],
                ui_pid, Pipeline.Rtsp.Edge.RENDERER_0, or0_ws_scene_dir, tr0_or0_scene_dir)

            return OrCreator([or0_args])
        else:
            return None

    @staticmethod
    def _create_speed(ui_pid, ui_entry_point, ui_identity_to_jnode_map, avaliable_ui_modules):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :type avaliable_ui_modules: set of UiIdentity
        :rtype: OrCreator | None
        """
        ui_spotspeed_rules_id_to_jnode_list = []
        ui_lane_enforcement_jnode = None
        ui_http_streaming_jnode = None
        ui_rtps_streaming_jnode = None

        for ui_module in avaliable_ui_modules:
            if ui_module.type == UI_TYPE_LANE_ENFORCEMENT:
                ui_lane_enforcement_jnode = ui_identity_to_jnode_map[ui_module]

            elif ui_module.type == UI_TYPE_HTTP_STREAMING:
                ui_http_streaming_jnode = ui_identity_to_jnode_map[ui_module]

            elif ui_module.type == UI_TYPE_RTSP_STREAMING:
                ui_rtps_streaming_jnode = ui_identity_to_jnode_map[ui_module]

            elif ui_module.type == UI_TYPE_SPOT_SPEED:
                ui_spotspeed_rules_id_to_jnode_list.append((ui_module.id, ui_identity_to_jnode_map[ui_module]))

        speed_limits_jnode = None
        if len(ui_spotspeed_rules_id_to_jnode_list) == 1:
            rule_id, spot_speed_jnode = ui_spotspeed_rules_id_to_jnode_list[0]
            rules_jnode = spot_speed_jnode['rw']['args']['rules']
            if len(rules_jnode) == 1:
                rule_jnode = rules_jnode[0]
                if rule_jnode['enabled']:
                    speed_limits_jnode = rule_jnode['limits']

        spool_dir = get_spool_dir(ui_pid)
        if ui_entry_point.type in [UI_TYPE_LANE_ENFORCEMENT, UI_TYPE_SPOT_SPEED,
                                   UI_TYPE_PIPELINE, UI_TYPE_GEOLOCATION,
                                   UI_TYPE_EVENTSTORE]:
            or0_ws_scene_dir = os.path.join(
                spool_dir,
                'or%s-ws%s-scene/' % (Pipeline.Speed.Edge.RENDERER_0.gid, Pipeline.Speed.Edge.WEBSERVER_0.gid))

            tr0_or0_scene_dir = os.path.join(
                spool_dir,
                'tr%s-or%s-scene/' % (Pipeline.Speed.Edge.TRACKER_0.gid, Pipeline.Speed.Edge.RENDERER_0.gid))

            or1_rs0_scene_dir = os.path.join(
                spool_dir,
                'or%s-rs%s-scene/' % (Pipeline.Speed.Edge.RENDERER_1.gid, Pipeline.Speed.Edge.RTSPSERVER_0.gid))

            tr0_or1_scene_dir = os.path.join(
                spool_dir,
                'tr%s-or%s-scene/' % (Pipeline.Speed.Edge.TRACKER_0.gid, Pipeline.Speed.Edge.RENDERER_1.gid))

            or0_args = OrCreator._serialize(
                ui_http_streaming_jnode,
                speed_limits_jnode,
                ui_lane_enforcement_jnode,
                ui_pid, Pipeline.Speed.Edge.RENDERER_0, or0_ws_scene_dir, tr0_or0_scene_dir)

            or1_args = OrCreator._serialize(
                ui_rtps_streaming_jnode,
                speed_limits_jnode,
                ui_lane_enforcement_jnode,
                ui_pid, Pipeline.Speed.Edge.RENDERER_1, or1_rs0_scene_dir, tr0_or1_scene_dir)

            return OrCreator([or0_args, or1_args])

        elif ui_entry_point.type in [UI_TYPE_HTTP_STREAMING]:
            or0_ws_scene_dir = os.path.join(
                spool_dir,
                'or%s-ws%s-scene/' % (Pipeline.Speed.Edge.RENDERER_0.gid, Pipeline.Speed.Edge.WEBSERVER_0.gid))

            tr0_or0_scene_dir = os.path.join(
                spool_dir,
                'tr%s-or%s-scene/' % (Pipeline.Speed.Edge.TRACKER_0.gid, Pipeline.Speed.Edge.RENDERER_0.gid))

            or0_args = OrCreator._serialize(
                ui_http_streaming_jnode,
                speed_limits_jnode,
                ui_lane_enforcement_jnode,
                ui_pid, Pipeline.Speed.Edge.RENDERER_0, or0_ws_scene_dir, tr0_or0_scene_dir)

            return OrCreator([or0_args])

        elif ui_entry_point.type in [UI_TYPE_RTSP_STREAMING]:
            or1_rs0_scene_dir = os.path.join(
                spool_dir,
                'or%s-rs%s-scene/' % (Pipeline.Speed.Edge.RENDERER_1.gid, Pipeline.Speed.Edge.RTSPSERVER_0.gid))

            tr0_or1_scene_dir = os.path.join(
                spool_dir,
                'tr%s-or%s-scene/' % (Pipeline.Speed.Edge.TRACKER_0.gid, Pipeline.Speed.Edge.RENDERER_1.gid))

            or1_args = OrCreator._serialize(
                ui_rtps_streaming_jnode,
                speed_limits_jnode,
                ui_lane_enforcement_jnode,
                ui_pid, Pipeline.Speed.Edge.RENDERER_1, or1_rs0_scene_dir, tr0_or1_scene_dir)

            return OrCreator([or1_args])

        else:
            raise NotImplemented

    @staticmethod
    def _create_mv(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: OrCreator | None
        """
        spool_dir = get_spool_dir(ui_pid)
        if ui_entry_point in [Pipeline.Mv.Ui.HTTP_STREAMING_0]:
            or0_ws_scene_dir = os.path.join(
                spool_dir,
                'or%s-ws%s-scene/' % (Pipeline.Mv.Edge.RENDERER_0.gid, Pipeline.Mv.Edge.WEBSERVER_0.gid))

            tr0_or0_scene_dir = os.path.join(
                spool_dir,
                'tr%s-or%s-scene/' % (Pipeline.Mv.Edge.TRACKER_0.gid, Pipeline.Mv.Edge.RENDERER_0.gid))

            or0_args = OrCreator._serialize(
                ui_identity_to_jnode_map[Pipeline.Mv.Ui.HTTP_STREAMING_0], None,
                ui_identity_to_jnode_map[Pipeline.Mv.Ui.LANE_ENFORCEMENT_0],
                ui_pid, Pipeline.Mv.Edge.RENDERER_0, or0_ws_scene_dir, tr0_or0_scene_dir)

            return OrCreator([or0_args])

        elif ui_entry_point in [Pipeline.Mv.Ui.RTSP_STREAMING_0]:
            or1_rs0_scene_dir = os.path.join(
                spool_dir,
                'or%s-rs%s-scene/' % (Pipeline.Mv.Edge.RENDERER_1.gid, Pipeline.Mv.Edge.RTSPSERVER_0.gid))

            tr0_or1_scene_dir = os.path.join(
                spool_dir,
                'tr%s-or%s-scene/' % (Pipeline.Mv.Edge.TRACKER_0.gid, Pipeline.Mv.Edge.RENDERER_1.gid))

            or1_args = OrCreator._serialize(
                ui_identity_to_jnode_map[Pipeline.Mv.Ui.RTSP_STREAMING_0], None,
                ui_identity_to_jnode_map[Pipeline.Mv.Ui.LANE_ENFORCEMENT_0],
                ui_pid, Pipeline.Mv.Edge.RENDERER_1, or1_rs0_scene_dir, tr0_or1_scene_dir)

            return OrCreator([or1_args])

        elif ui_entry_point in [Pipeline.Mv.Ui.LANE_ENFORCEMENT_0, Pipeline.Mv.Ui.PIPELINE,
                                Pipeline.Mv.Ui.GEOLOCATION_0, Pipeline.Mv.Ui.EVENTSTORE_0]:
            or0_ws_scene_dir = os.path.join(
                spool_dir,
                'or%s-ws%s-scene/' % (Pipeline.Mv.Edge.RENDERER_0.gid, Pipeline.Mv.Edge.WEBSERVER_0.gid))

            tr0_or0_scene_dir = os.path.join(
                spool_dir,
                'tr%s-or%s-scene/' % (Pipeline.Mv.Edge.TRACKER_0.gid, Pipeline.Mv.Edge.RENDERER_0.gid))

            or1_rs0_scene_dir = os.path.join(
                spool_dir,
                'or%s-rs%s-scene/' % (Pipeline.Mv.Edge.RENDERER_1.gid, Pipeline.Mv.Edge.RTSPSERVER_0.gid))

            tr0_or1_scene_dir = os.path.join(
                spool_dir,
                'tr%s-or%s-scene/' % (Pipeline.Mv.Edge.TRACKER_0.gid, Pipeline.Mv.Edge.RENDERER_1.gid))

            or0_args = OrCreator._serialize(
                ui_identity_to_jnode_map[Pipeline.Mv.Ui.HTTP_STREAMING_0], None,
                ui_identity_to_jnode_map[Pipeline.Mv.Ui.LANE_ENFORCEMENT_0],
                ui_pid, Pipeline.Mv.Edge.RENDERER_0, or0_ws_scene_dir, tr0_or0_scene_dir)

            or1_args = OrCreator._serialize(
                ui_identity_to_jnode_map[Pipeline.Mv.Ui.RTSP_STREAMING_0], None,
                ui_identity_to_jnode_map[Pipeline.Mv.Ui.LANE_ENFORCEMENT_0],
                ui_pid, Pipeline.Mv.Edge.RENDERER_1, or1_rs0_scene_dir, tr0_or1_scene_dir)

            return OrCreator([or0_args, or1_args])

        else:
            raise NotImplemented

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map, avaliable_ui_modules):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :type avaliable_ui_modules: set of UiIdentity
        :rtype: OrCreator
        """
        if ui_entry_point.ptype == PTYPE_SPEED:
            creator = OrCreator._create_speed(ui_pid, ui_entry_point, ui_identity_to_jnode_map, avaliable_ui_modules)

        elif ui_entry_point.ptype == PTYPE_LPR_RTSP:
            creator = OrCreator._create_lpr_rtsp(ui_pid, ui_entry_point, ui_identity_to_jnode_map)

        elif ui_entry_point.ptype == PTYPE_LPR_MV:
            creator = OrCreator._create_mv(ui_pid, ui_entry_point, ui_identity_to_jnode_map)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_streaming_jnode, ui_speed_enforcement_jnode, ui_lane_enforcement_jnode, ui_pid, edge_identity,
                   scene_dir, source_dir):
        """
        :type ui_streaming_jnode: dict
        :type ui_speed_enforcement_jnode: dict | None
        :type ui_lane_enforcement_jnode: dict
        :type ui_pid: str
        :type edge_identity: EdgeIdentity
        :type scene_dir: str
        :type source_dir: str
        :return: str, dict, str
        """
        codec = "JPEG"
        if edge_identity in [Pipeline.Mv.Edge.RENDERER_1, Pipeline.Speed.Edge.RENDERER_1]:
            codec = "RGBA"

        ui_live_jnode = ui_streaming_jnode['rw']['live']['videostream']
        ui_vehicle_jnode = ui_streaming_jnode['rw']['live']['renderer']['vehicle']

        image_w = ui_live_jnode['frame-w']
        main_conf_jnode = {
            "Source": {
                "Url": source_dir
            },
            "services": {
                'origin': get_origin_about_filepath(ui_pid),
                "datetime": UI_TIME_JNODE_FILEPATH,
                "geolocation": "/etc/edge/setupstore/admin/actions/pipeline/%s/geolocation/0.json" % ui_pid,
                "monitoring": {
                    "path": "/var/spool/edge/pipeline-%s/zabbix/or%s" % (ui_pid, edge_identity.gid),
                    "tout": 30
                }
            },
            "Renderer": {
                "profile": {
                    "width": image_w,
                    "annotation": {
                        "vehicle": {
                            "lr": {
                                "plate": ui_vehicle_jnode.get('video-plate', 'never'),
                                "dir": ui_vehicle_jnode.get('video-dir', 'never'),
                                "lane": ui_vehicle_jnode.get('video-lane', 'never'),
                                "trace": ui_vehicle_jnode.get('video-plate-trace', 'never'),
                                "body": ui_vehicle_jnode.get('video-body-image', 'never'),
                                "center": ui_vehicle_jnode.get('video-plate-center', 'never')
                            },
                            "ra": {
                                "pos": ui_vehicle_jnode.get('radar-position', 'never'),
                                "speed": ui_vehicle_jnode.get('radar-speed', 'never'),
                                "len": ui_vehicle_jnode.get('radar-length', 'never'),
                                "dir": ui_vehicle_jnode.get('radar-dir', 'never')
                            }
                        }
                    },
                    "footer": ui_streaming_jnode['rw']['live']['footer']
                }
            },
            "image-store": {
                "format": codec,
                "endpoint": "/var/spool/edge/pipeline-%s/or%s-or%s-store/" % (ui_pid, edge_identity.gid, edge_identity.gid),
                "max-count": 8
            },
            "scene-event": {
                "max-file-count": 8,
                "endpoints": [scene_dir]
            }
        }
        if ui_speed_enforcement_jnode is not None:
            speed_limit = ui_speed_enforcement_jnode['speed-limit']
            main_conf_jnode['Renderer']['YellowSpeedThreshold'] = speed_limit
            main_conf_jnode['Renderer']['RedSpeedThreshold'] = speed_limit + ui_speed_enforcement_jnode['max-overspeed']

        if len(ui_lane_enforcement_jnode['rw']['args']['lanes']) > 0 and ui_vehicle_jnode['video-lane'] != 'never':
            main_conf_jnode['road-mask'] = {
                'lanes': ui_lane_enforcement_jnode['rw']['args']['lanes'],
                'scene-wh': ui_lane_enforcement_jnode['rw']['args']['scene-wh']
            }

        edge_group = helpers.get_edge_group(ui_pid)
        edge_id = helpers.get_edge_id(edge_group, edge_identity)

        if ui_streaming_jnode['rw']['enabled']:
            setupstore_conf_text = get_setupstore_conf_text(edge_id)
        else:
            setupstore_conf_text = get_setupstore_conf_text(edge_id, active=False)

        return edge_id, main_conf_jnode, setupstore_conf_text

    def serialize(self):
        for (edge_id, main_conf_jnode, setup_store_text) in self._id_main_setupstore_list:
            edge_jnode_dirpath = get_edge_jnode_dirpath(EDGE_TYPE_RENDERER, edge_id)
            if not os.path.exists(edge_jnode_dirpath):
                helpers.silent_mkdir_r(edge_jnode_dirpath)

            main_conf_filepath = os.path.join(edge_jnode_dirpath, 'main.conf')
            with open(main_conf_filepath, 'w') as json_file:
                ujson.dump(main_conf_jnode, json_file, indent=2)

            setupstore_conf_filepath = os.path.join(edge_jnode_dirpath, 'setupstore.conf')
            with open(setupstore_conf_filepath, 'w') as text_file:
                text_file.write(setup_store_text)


class OrDeleter:
    def __init__(self, edge_rn_dirpath_list):
        """
        :type edge_rn_dirpath_list: list of str
        """
        self._edge_rn_dirpath_list = edge_rn_dirpath_list

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: OrDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            or0_id = helpers.get_edge_id(edge_group, Pipeline.Rtsp.Edge.RENDERER_0)
            or0_type = Pipeline.Rtsp.Edge.RENDERER_0.type
            deleter = OrDeleter([get_edge_jnode_dirpath(or0_type, or0_id)])
            return deleter

        elif ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            or0_id = helpers.get_edge_id(edge_group, Pipeline.Mv.Edge.RENDERER_0)
            or0_type = Pipeline.Mv.Edge.RENDERER_0.type

            or1_id = helpers.get_edge_id(edge_group, Pipeline.Mv.Edge.RENDERER_1)
            or1_type = Pipeline.Mv.Edge.RENDERER_1.type
            deleter = OrDeleter([get_edge_jnode_dirpath(or0_type, or0_id), get_edge_jnode_dirpath(or1_type, or1_id)])
            return deleter

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            or0_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.RENDERER_0)
            or0_type = Pipeline.Speed.Edge.RENDERER_0.type

            or1_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.RENDERER_1)
            or1_type = Pipeline.Speed.Edge.RENDERER_1.type
            deleter = OrDeleter([get_edge_jnode_dirpath(or0_type, or0_id), get_edge_jnode_dirpath(or1_type, or1_id)])
            return deleter

        else:
            raise NotImplemented

    def run(self):
        for edge_rn_dirpath in self._edge_rn_dirpath_list:
            helpers.silent_rm_r(edge_rn_dirpath)