import ujson
import os

from edge.admin.ui_types import BaseCreator
from edge.path_helpers import *
from os.path import join as join_path

import edge.pipeline_helpers as helpers


def rs_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :type avaliable_ui_modules: set of UiIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    # speed
    if ui_entry_point in [Pipeline.Speed.Ui.RTSP_STREAMING_0, Pipeline.Speed.Ui.VIDEOSOURCE_0]:
        used_ui_modules.update([Pipeline.Speed.Ui.RTSP_STREAMING_0, Pipeline.Speed.Ui.VIDEOSOURCE_0])
        used_edge_modules.add(Pipeline.Speed.Edge.RTSPSERVER_0)
        creator_ctor = RsCreator.create

    elif ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Speed.Ui.RTSP_STREAMING_0, Pipeline.Speed.Ui.VIDEOSOURCE_0])
        creator_ctor = RsCreator.create

    # mv
    elif ui_entry_point in [Pipeline.Mv.Ui.RTSP_STREAMING_0, Pipeline.Mv.Ui.VIDEOSOURCE_0]:
        used_ui_modules.update([Pipeline.Mv.Ui.RTSP_STREAMING_0, Pipeline.Mv.Ui.VIDEOSOURCE_0])
        used_edge_modules.add(Pipeline.Mv.Edge.RTSPSERVER_0)
        creator_ctor = RsCreator.create

    elif ui_entry_point in [Pipeline.Mv.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Mv.Ui.RTSP_STREAMING_0, Pipeline.Mv.Ui.VIDEOSOURCE_0])
        creator_ctor = RsCreator.create

    return creator_ctor


class RsCreator(BaseCreator):
    def __init__(self, id_main_setupstore):
        """
        :type id_main_setupstore: str, dict, str
        """
        BaseCreator.__init__(self)
        self._id_main_setupstore = id_main_setupstore

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: RsCreator
        """
        if ui_entry_point in [Pipeline.Speed.Ui.RTSP_STREAMING_0, Pipeline.Speed.Ui.PIPELINE,
                              Pipeline.Speed.Ui.VIDEOSOURCE_0, Pipeline.Mv.Ui.RTSP_STREAMING_0,
                              Pipeline.Mv.Ui.PIPELINE, Pipeline.Mv.Ui.VIDEOSOURCE_0]:
            id_main_setupstore = RsCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map)
            creator = RsCreator(id_main_setupstore)

        else:
            raise NotImplemented('entry point not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: str, dict, str
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point.ptype == PTYPE_SPEED:
            ui_rtsp_streaming_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.RTSP_STREAMING_0]
            ui_videosource_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.VIDEOSOURCE_0]
            rtsp_server_gid = Pipeline.Speed.Edge.RTSPSERVER_0.gid
            overlay_gid = Pipeline.Speed.Edge.RENDERER_1.gid
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.RTSPSERVER_0)

        elif ui_entry_point.ptype == PTYPE_LPR_MV:
            ui_rtsp_streaming_jnode = ui_identity_to_jnode_map[Pipeline.Mv.Ui.RTSP_STREAMING_0]
            ui_videosource_jnode = ui_identity_to_jnode_map[Pipeline.Mv.Ui.VIDEOSOURCE_0]
            rtsp_server_gid = Pipeline.Mv.Edge.RTSPSERVER_0.gid
            overlay_gid = Pipeline.Mv.Edge.RENDERER_1.gid
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Mv.Edge.RTSPSERVER_0)

        else:
            raise NotImplemented

        # get main.conf#/Resolution
        ui_videostream_jnode = ui_rtsp_streaming_jnode['rw']['live']['videostream']

        # get main.conf#/Stream/UdpPorts
        ui_streaming_jnode = ui_rtsp_streaming_jnode['rw']['streaming']
        if ui_streaming_jnode['transport'] == 'rtp-over-ucast-udp-or-tcp':
            if ui_streaming_jnode['min-udp-port'] < ui_streaming_jnode['max-udp-port']:
                edge_udp_ports = ui_streaming_jnode['min-udp-port'], ui_streaming_jnode['max-udp-port']
            else:
                edge_udp_ports = ui_streaming_jnode['max-udp-port'], ui_streaming_jnode['min-udp-port']

        elif ui_streaming_jnode['transport'] == 'rtp-over-tcp':
            edge_udp_ports = None, None

        else:
            raise NotImplemented

        # get main.conf#/Stream/Link and main.conf#/Quality
        ui_tcp_port = ui_streaming_jnode['tcp-port']
        if ui_videostream_jnode['format'] == "h264-cqp":
            link = 'rtsp://@:%s/live-h264' % ui_tcp_port
            quality = ui_videostream_jnode['format-h264-cqp-compression']

        elif ui_videostream_jnode['format'] == "h264-cbr":
            link = 'rtsp://@:%s/live-h264' % ui_tcp_port
            quality = ui_videostream_jnode['format-h264-cbr-bitrate-max']

        elif ui_videostream_jnode['format'] == "mjpeg":
            link = 'rtsp://@:%s/live-mjpeg' % ui_tcp_port
            quality = ui_videostream_jnode['format-mjpeg-quality']

        else:
            raise NotImplemented

        spool_dir = get_spool_dir(ui_pid)

        main_conf = {
            "Stream": {
                "Link": link,
                "UdpPorts": edge_udp_ports
            },
            "UseGPU": not ui_videostream_jnode['sw-codec'],
            "Quality": quality,
            "FileApi": {
                "AddSceneImgReact": {
                    "Url": join_path(spool_dir, 'or%s-rs%s-scene' % (overlay_gid, rtsp_server_gid))
                }
            },
            "Fps": ui_videosource_jnode['rw']['videostream']['fps'] // ui_videostream_jnode['skip-rate'],
            "Resolution": {
                "MaxHeight": 65535,
                "MaxWidth": ui_videostream_jnode['frame-w']
            }
        }
        # create setupstore.conf
        if ui_rtsp_streaming_jnode['rw']['enabled']:
            setupstore_conf_text = get_setupstore_conf_text(edge_id)
        else:
            setupstore_conf_text = get_setupstore_conf_text(edge_id, active=False)

        return edge_id, main_conf, setupstore_conf_text

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text = self._id_main_setupstore

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_RTSP_SERVER, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_abspath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_abspath, 'w') as main_conf:
            ujson.dump(main_conf_jnode, main_conf, indent=2)

        setupstore_conf_path = join_path(edge_jnode_dir, 'setupstore.conf')

        with open(setupstore_conf_path, 'w') as text_file:
            text_file.write(setupstore_conf_text)


class RsDeleter:
    def __init__(self, edge_rs_dirpath):
        """
        :type edge_rs_dirpath: str
        """
        self._edge_rs_dirpath = edge_rs_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: RsDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.RTSPSERVER_0)
            edge_type = Pipeline.Speed.Edge.RTSPSERVER_0.type
            deleter = RsDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Mv.Edge.RTSPSERVER_0)
            edge_type = Pipeline.Mv.Edge.RTSPSERVER_0.type
            deleter = RsDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            return None

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_rs_dirpath)
