import jsonschema
import ujson

from edge import eventlog, path_helpers as ph, pipeline_helpers as oh
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.types import FormatTemplate
from edge.com import fs

import edge_creator as creator

from ui_pipeline_set import get_pipeline_type, add_detector_to_pipeline, delete_detector_from_pipeline, rename_detector


def _validate_format(jnode):
    """
    :type jnode: list
    :rtype: bool
    """
    key_to_true_map = {
        'speed': True,
        'datetime': True,
        'datetime-ms': True,
        'post-name': True,
        'geolocation': True,
        'plate': True,
        'lane': True,
        'dir': True,
        'video-speed': True,
        'video-speed2': True,
        'radar-speed': True,
        'radar-speed2': True,
        'max-speed': True,
        'radar-overspeed': True,
        'vendor': True,
        'model': True,
        'serial': True
    }

    for row in jnode:
        for col in row['cols']:
            try:
                FormatTemplate(template=col['format']).substitute(key_to_true_map)
            except KeyError as x:
                eventlog.log_d('fail: invalid key (%s)' % x.message)
                return False

    return True


class UiSpotSpeed:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """
        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def _handle_get(handler_args):
        """
        :type handler_args: (str, str)
        :rtype: dict
        """
        ui_pid, ui_jnode_filepath = handler_args
        ui_jnode = fs.load_json2(ui_jnode_filepath)

        ui_jnode['rw']['limits'] = oh.get_detector_limits(ui_pid=ui_pid, videosource_ui_id='0', lanes=True)

        return rh.reply_done(ui_jnode['rw'])

    @staticmethod
    def _handle_put(handler_args):
        """
        :type handler_args: (srt, UiIdentity, dict)
        :rtype: dict
        """
        ui_pid, entry_point, new_ui_jnode = handler_args

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
                return False

            if new_ui_jnode['args']['name'] != ui_identity_to_jnode_map[entry_point]['rw']['args']['name']:
                rename_detector(ui_pid, entry_point.type, entry_point.id, new_ui_jnode['args']['name'])

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
            return True

        try:
            creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            new_ui_jnode['limits'] = oh.get_detector_limits(ui_pid=ui_pid, videosource_ui_id='0', lanes=True)
            return rh.reply_done(new_ui_jnode)

        except ui.LockError:
            return rh.reply_fail(rh.REPLY_CONFLICT_409)

        except Exception as x:
            eventlog.log_d("fail: %s" % x.message)
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _handle_post(handler_args):
        """
        :type handler_args: (str, str, str)
        """
        ui_pid, ui_type, name = handler_args
        # add new module pipeline.json file
        result = add_detector_to_pipeline(ui_pid, ui_type, name)
        if result:
            new_ui_id, new_module_list = result
            ui_ptype = get_pipeline_type(ui_pid)

            if ui_ptype is None:
                return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_NOT_FOUND_404)

            entry_point = ph.UiIdentity(ui_ptype, ui_type, new_ui_id)

            # create ui jnode from default
            new_ui_jnode_filepath = creator.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, new_ui_id)
            def_jnode_filepath = ph.get_ui_def_jnode_filepath(entry_point, None)
            fs.create_jnode_from_default(def_jnode_filepath, new_ui_jnode_filepath)

            # create schema symlink
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, entry_point.type, new_ui_id)
            fs.ln_schema(ui_schema_filepath, new_ui_jnode_filepath)

            def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
                """
                :type ui_identity_to_jnode_map: dict of UiIdentity: dict
                :rtype: bool
                """
                ui_identity_to_jnode_map[entry_point]['rw']['args']['name'] = name
                return True

            try:
                creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)
                return rh.reply_done(new_module_list)

            except ui.LockError:
                return rh.reply_fail(rh.REPLY_CONFLICT_409)

            except Exception as x:
                eventlog.log_d("fail: %s" % x.message)
                return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

        else:
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _handle_delete(handler_args):
        """
        :type handler_args: (str, UiIdentity)
        """
        ui_pid, entry_point = handler_args

        # remove ui_jnode from disk
        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        fs.silent_rm_r(ui_jnode_filepath)

        # remove ui_schema from disk
        ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, entry_point.type, entry_point.id)
        fs.silent_rm_r(ui_schema_filepath)

        # remove module from pipeline.json
        res = delete_detector_from_pipeline(ui_pid, entry_point.type, entry_point.id, '')

        try:
            creator.generate(ui_pid, entry_point, None)
            return rh.reply_done(res)

        except ui.LockError:
            return rh.reply_fail(rh.REPLY_CONFLICT_409)

        except Exception as x:
            eventlog.log_d("fail: %s" % x.message)
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _deserialize_post(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiSpotSpeed
        """
        if query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
        except IndexError:
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_NOT_FOUND_404)

        args = ujson.loads(query_args.raw_body)
        ui_type = args.get('type')
        det_name = args.get('name')

        if ui_type != ph.UI_TYPE_SPOT_SPEED or det_name is None:
            eventlog.log_d('fail: invalid params [%s]' % query_args.raw_body)
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        handler_args = ui_pid, ui_type, det_name
        return UiSpotSpeed(UiSpotSpeed._handle_post, handler_args)

    @staticmethod
    def _deserialize_delete(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiSpotSpeed
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_d('fail: bad-request')
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)

        if ui_ptype is None:
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_NOT_FOUND_404)

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        handler_args = ui_pid, entry_point
        return UiSpotSpeed(UiSpotSpeed._handle_delete, handler_args)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiSpotSpeed
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_d('fail: bad-request')
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_NOT_FOUND_404)

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)

        return UiSpotSpeed(UiSpotSpeed._handle_get, (ui_pid, ui_jnode_filepath))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiSpotSpeed
        """
        try:
            ui_pid = query_args.get_path_info_item(ui.UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(ui.UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(ui.UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)

        if ui_ptype is None:
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_NOT_FOUND_404)

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        # load new-ui-jnode
        new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_d('fail: bad-request')
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        # load ui-schena-jnode
        try:
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_d("fail: load ui-schema-jnode (%s)" % x.message)
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_INTERNAL_SERVER_ERROR_500)

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_d("fail: validate new-ui-jnode (%s)" % str(x.message))
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        # validate "format" jnodes from json
        images_jnode = new_ui_jnode['args']['images']['violation']
        vehicle_body_footer_rows = images_jnode['vehicle']['footer']['rows']
        scene_alt_footer_rows = images_jnode['scene']['footer']['rows']

        if _validate_format(vehicle_body_footer_rows) is False or _validate_format(scene_alt_footer_rows) is False:
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        # all arguments are valid
        return UiSpotSpeed(UiSpotSpeed._handle_put, (ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiSpotSpeed
        """
        if query_args.http_method == 'GET':
            return UiSpotSpeed._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiSpotSpeed._deserialize_put(query_args)
        
        elif query_args.http_method == 'POST':
            return UiSpotSpeed._deserialize_post(query_args)

        elif query_args.http_method == 'DELETE':
            return UiSpotSpeed._deserialize_delete(query_args)

        else:
            eventlog.log_d("fail: method-not-allowed")
            return UiSpotSpeed(handler=rh.reply_fail, handler_args=rh.REPLY_METHOD_NOT_ALLOWED_405)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
