import os

from edge import eventlog
from edge import path_helpers as ph
from edge import pipeline_helpers as oh
from edge.admin.reply_helpers import (
    Reply, REPLY_NOT_FOUND_404, REPLY_SUCCESS_200,
    REPLY_BAD_REQUEST_400, REPLY_INTERNAL_SERVER_ERROR_500)
from edge.admin.cgi import (
    run_cgi, CgiModule, CgiPathInfo, CgiRequest,
    CgiBadRequestError, CgiNotFoundError)
from setupstore.actions.pipeline.edge_creator import generate, LockError

from _api import get, put, put_crt_key


def _handle_regenerate():
    pipeline_jnode_set = oh.get_pipeline_jnode_set2()
    if pipeline_jnode_set is None:
        eventlog.log_d("fail: pipeline-jnode-set is absent")
        return

    for pipeline_jnode in pipeline_jnode_set:
        ui_pid = pipeline_jnode['id']
        ui_ptype = pipeline_jnode['type']

        eventlog.log_d('try: pipeline-%s' % ui_pid)
        # noinspection PyBroadException
        try:
            # create edge-jnode-set
            generate(ui_pid, ph.UiIdentity(ui_ptype, ph.UI_TYPE_SIGN, '0'), None)
            eventlog.log_d('done: pipeline-%s' % ui_pid)
            continue

        except NotImplementedError:
            eventlog.log_d('fail: pipeline-%s (camera not supported)' % ui_pid)

        except LockError:
            eventlog.log_d('fail: pipeline-%s (lock error)' % ui_pid)

        except Exception:
            eventlog.log_d('fail: pipeline-%s (traceback)' % ui_pid)
            eventlog.log_x()

    eventlog.log_d('regeneration config done')


class _PathInfo(CgiPathInfo):
    def __init__(self, *args, **kwargs):
        CgiPathInfo.__init__(self, *args, **kwargs)

    @property
    def file_type(self):
        return self[0]

    @staticmethod
    def unpack(raw_path_info):
        """
        :type raw_path_info: str
        :rtype: _PathInfo
        """
        # eventlog.log_d('unpack %s' % type(raw_path_info))
        return _PathInfo(raw_path_info.split('/')[1:])


class _Module(CgiModule):
    """
    GET     /api/v2/admin/actions/sign
    PUT     /api/v2/admin/actions/sign
    PUT     /api/v2/admin/actions/sign/install
    """
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request):
        """
        :type request: CgiRequest
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._get,
            ],
            'PUT': [
                cls._put,
                cls._put_install
            ],
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except CgiBadRequestError:
                raise

            except CgiNotFoundError:
                pass

        eventlog.log_i('fail: http404')
        raise CgiNotFoundError

    @staticmethod
    def _get(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """

            result = get()
            rc = REPLY_SUCCESS_200 if result else REPLY_NOT_FOUND_404
            return Reply(status=rc, body=result)

        return _Module(handler)

    @staticmethod
    def _put(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body is None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            result = put(request.body_jnode)

            status = REPLY_NOT_FOUND_404
            if result:
                status = REPLY_SUCCESS_200

                # regenerate speedcontrol configs
                _handle_regenerate()

            return Reply(status=status, body=result)

        return _Module(handler)

    @staticmethod
    def _put_install(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 1:
                raise CgiNotFoundError

            if path_info.file_type != "install":
                raise CgiNotFoundError

        def check_args():
            """
            :rtype: (str, str, str | None)
            """
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            crt_path = os.environ.get('CRT_PATH')
            if crt_path is None:
                eventlog.log_e('fail: http400 (env.CRT_PATH not found)')
                raise CgiBadRequestError

            key_path = os.environ.get('KEY_PATH')
            if key_path is None:
                eventlog.log_e('fail: http400 (env.KEY_PATH not found)')
                raise CgiBadRequestError

            key_pass = os.environ.get('KEY_PASS', None)

            return crt_path, key_path, key_pass

        check_path_info(request.path_info)
        crt_filepath, key_filepath, key_password = check_args()

        def handler():
            """
            :rtype: Reply
            """
            reply = put_crt_key(crt_filepath, key_filepath, key_password)

            if reply == 500:
                status = REPLY_INTERNAL_SERVER_ERROR_500
                reply = None
            elif reply == 400:
                status = REPLY_BAD_REQUEST_400
                reply = None
            else:
                status = REPLY_SUCCESS_200

                # restart speedcontrol
                oh.restart_edge_modules(ph.EDGE_TYPE_SPEEDCONTROL)

            return Reply(status=status, body=reply)

        return _Module(handler)


def _main():
    run_cgi(
        lambda arg: _PathInfo.unpack(arg),
        lambda arg: _Module.unpack(arg))


if __name__ == '__main__':
    eventlog.open_log('cgi-sign', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
