#!/usr/bin/python
import os
import shutil
import time
import uuid
import ujson
import sys

from edge import eventlog
import jsonschema
from edge.admin.reply_helpers import *
from edge.com.fs import ln_schema2, silent_mkdir_r, try_lock_pid_file

UI_VPN_SCHEMA_FILEPATH = '/etc/edge/setupstore/schemas/vpn-profile.json'
UI_VPN_DEFAULT_SCHEMA_FILEPATH = '/opt/edge/share/setupstore/ui/vpn-profile-schema.json'
SETUPSTORE_VPN_PROFILE_DIRPATH = '/etc/edge/setupstore/admin/actions/vpn-profile/'
LOCK_FILEPATH = '/var/run/edge/vpn-profile'

ETC_DEFAULT_OPENVPN_FILEPATH = '/etc/default/openvpn'
ETC_OPENVPN_DIRPATH = '/etc/openvpn/'

RESTART_NONE = 0
RESTART_VPN = 1

BACKUP_DIRPATH = '/etc/edge/setupstore/backup/vpn-profile.' + \
                 str(round(time.time() * 1000000)) + '/'


def try_create_backup(profile_guid):
    # noinspection PyBroadException
    try:
        os.makedirs(BACKUP_DIRPATH)
        if not os.path.exists(SETUPSTORE_VPN_PROFILE_DIRPATH):
            os.makedirs(SETUPSTORE_VPN_PROFILE_DIRPATH)
        shutil.copy(SETUPSTORE_VPN_PROFILE_DIRPATH + profile_guid + '.json', BACKUP_DIRPATH)
        shutil.copy(ETC_DEFAULT_OPENVPN_FILEPATH, BACKUP_DIRPATH)
        shutil.copy(ETC_OPENVPN_DIRPATH + profile_guid + '.conf', BACKUP_DIRPATH)
        return True

    except Exception:
        pass

    return False


def try_restore_backup(profile_guid):
    # noinspection PyBroadException
    try:
        shutil.copy(BACKUP_DIRPATH + profile_guid + '.json', SETUPSTORE_VPN_PROFILE_DIRPATH + profile_guid + '.json')
        shutil.copy(BACKUP_DIRPATH + 'default_openvpn', ETC_DEFAULT_OPENVPN_FILEPATH)
        shutil.copy(BACKUP_DIRPATH + profile_guid + '.conf', ETC_OPENVPN_DIRPATH + profile_guid + '.conf')
        os.rmdir(BACKUP_DIRPATH)
        return True

    except Exception:
        return False


def remove_backup():
    # noinspection PyBroadException
    try:
        shutil.rmtree(BACKUP_DIRPATH)
    except Exception:
        pass


def _check_profile_changed(old_jnode, new_jnode):
    """
    :type old_jnode: dict
    :type new_jnode: dict
    :rtype: bool
    """
    if old_jnode['id'] != new_jnode['id']:
        return True
    if old_jnode['name'] != new_jnode['name']:
        return True
    if old_jnode['active'] != new_jnode['active']:
        return True
    if old_jnode['interface'] != new_jnode['interface']:
        return True
    if old_jnode['server-host'] != new_jnode['server-host']:
        return True
    if old_jnode['server-port'] != new_jnode['server-port']:
        return True
    if old_jnode['protocol'] != new_jnode['protocol']:
        return True
    if old_jnode['ca-cert'] != new_jnode['ca-cert']:
        return True

    if old_jnode['client-cert'] != new_jnode['client-cert']:
        return True
    if old_jnode['client-key'] != new_jnode['client-key']:
        return True
    if old_jnode['mtu-size'] != new_jnode['mtu-size']:
        return True
    if old_jnode['cipher'] != new_jnode['cipher']:
        return True
    if old_jnode['hmac'] != new_jnode['hmac']:
        return True
    if old_jnode['internal'] != new_jnode['internal']:
        return True

    return False


def write_text_file(file_abspath, file_content):
    with open(file_abspath, 'w') as destination_file:
        destination_file.write(file_content)


def get_profile_jnode(profile_guid):
    """
    :type profile_guid: str
    :rtype: dict | None
    """
    profile_jnode_abspath = SETUPSTORE_VPN_PROFILE_DIRPATH + profile_guid + '.json'
    if os.path.exists(profile_jnode_abspath):
        with open(profile_jnode_abspath, 'r') as old_jnode_file:
            profile_jnode = ujson.load(old_jnode_file)
            return profile_jnode

    return None


def edit_certificates(profile_guid, new_jnode):
    ca_cert_abspath = ETC_OPENVPN_DIRPATH + profile_guid + '.ca'
    client_cert_abspath = ETC_OPENVPN_DIRPATH + profile_guid + '.cert'
    client_key_abspath = ETC_OPENVPN_DIRPATH + profile_guid + '.key'

    ca_cert_content = new_jnode['ca-cert']
    client_cert_content = new_jnode['client-cert']
    client_key_content = new_jnode['client-key']

    #
    if ca_cert_content:
        write_text_file(ca_cert_abspath, ca_cert_content)

    elif not ca_cert_content and os.path.exists(ca_cert_abspath):
        old_jnode = get_profile_jnode(profile_guid)
        ca_cert_content = old_jnode['ca-cert']
        write_text_file(ca_cert_abspath, ca_cert_content)

    else:
        ca_cert_abspath = None

    #
    if client_cert_content:
        write_text_file(client_cert_abspath, client_cert_content)

    elif not client_cert_content and os.path.exists(client_cert_abspath):
        old_jnode = get_profile_jnode(profile_guid)
        client_cert_content = old_jnode['client-cert']
        write_text_file(client_cert_abspath, client_cert_content)

    else:
        client_cert_abspath = None

    #
    if client_key_content:
        write_text_file(client_key_abspath, client_key_content)

    elif not client_key_content and os.path.exists(client_key_abspath):
        old_jnode = get_profile_jnode(profile_guid)
        client_key_content = old_jnode['client-key']
        write_text_file(client_key_abspath, client_key_content)

    else:
        client_key_abspath = None

    return ca_cert_abspath, client_cert_abspath, client_key_abspath


def edit_vpn_profile_conf(profile_guid, new_jnode):
    """
    :type profile_guid: str
    :type new_jnode: dict
    :return:
    """
    conf_lines = [
        '\n# mandatory\nclient\nuser nobody\ngroup nogroup\nverb 3\n'
        '#log /var/log/openvpn.log\nstatus /var/log/openvpn-status.log\n#'
    ]

    interface = new_jnode['interface']
    conf_lines.append('\ndev %s\n' % interface)

    if new_jnode['compression']:
        conf_lines.append('comp-lzo\n')

    server_host = new_jnode['server-host']
    server_port = new_jnode['server-port']
    conf_lines.append('remote %s %s\n' % (server_host, server_port))

    prototcol = new_jnode['protocol']
    conf_lines.append('\nproto %s\n' % prototcol)

    ca_cert_abspath, client_cert_abspath, client_key_abspath = edit_certificates(profile_guid, new_jnode)

    if ca_cert_abspath is not None:
        conf_lines.append('ca %s\n' % ca_cert_abspath)

    if client_cert_abspath is not None:
        conf_lines.append('cert %s\n' % client_cert_abspath)

    if client_key_abspath is not None:
        conf_lines.append('key %s\n' % client_key_abspath)

    mtu_size = new_jnode['mtu-size']
    cipher = new_jnode['cipher']
    hmac = new_jnode['hmac']
    conf_lines.append('tun-mtu %s\ncipher %s\nauth %s\n' % (mtu_size, cipher, hmac))

    profile_conf_abspath = ETC_OPENVPN_DIRPATH + profile_guid + '.conf'
    with open(profile_conf_abspath, 'w') as conf_file:
        conf_file.write(''.join(conf_lines))


def get_active_profile_guid():
    """
    :rtype: std | None
    """
    with open(ETC_DEFAULT_OPENVPN_FILEPATH, 'r') as old_default_vpn_file:
        old_defaults_vpn_lines = old_default_vpn_file.readlines()

    for old_line in old_defaults_vpn_lines:
        if old_line.startswith('AUTOSTART='):
            profile_guid = old_line.split('=')[1].replace('"', '').strip()

            return profile_guid

    return None


def set_autostart_profile(new_active_profile_guid):
    """
    :type new_active_profile_guid: str | None
    :rtype: int
    """
    # Edit profile autostart.
    with open(ETC_DEFAULT_OPENVPN_FILEPATH, 'r') as old_default_vpn_file:
        old_defaults_vpn_lines = old_default_vpn_file.readlines()

    new_defaults_vpn_lines = []
    for old_line in old_defaults_vpn_lines:
        if old_line.startswith('#'):
            new_defaults_vpn_lines.append(old_line)

    if new_active_profile_guid is None:
        autostart_line = 'AUTOSTART="none"\n'

    else:
        autostart_line = 'AUTOSTART="%s"\n' % new_active_profile_guid

    new_defaults_vpn_lines.append(autostart_line)

    with open(ETC_DEFAULT_OPENVPN_FILEPATH, 'w') as new_default_vpn_file:
        new_default_vpn_file.write(''.join(new_defaults_vpn_lines))

    return RESTART_VPN


def handle_put_profile(profile_guid, old_jnode, new_jnode):
    if not try_lock_pid_file(LOCK_FILEPATH):
        return reply_fail(REPLY_CONFLICT_409)

    if old_jnode['internal']:
        return reply_fail(REPLY_BAD_REQUEST_400)

    if not try_create_backup(profile_guid):
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    error = False
    if _check_profile_changed(old_jnode, new_jnode):
        try:
            edit_vpn_profile_conf(profile_guid, new_jnode)

            old_active_profile_guid = get_active_profile_guid()
            if new_jnode['active']:
                # Activate another profile.
                set_autostart_profile(profile_guid)

            elif old_active_profile_guid == profile_guid:
                # Deactivate current profile.
                set_autostart_profile(profile_guid if new_jnode['active'] else None)

            else:
                # Don't change current active profile's autostart settings.
                pass

            with open(SETUPSTORE_VPN_PROFILE_DIRPATH + profile_guid + '.json', 'w') as f:
                f.write((ujson.dumps(new_jnode, indent=4)))

            _openvpn_reload_restart()

        except Exception:
            error = True

    if not error:
        remove_backup()
        return reply_done(new_jnode)

    if try_restore_backup(profile_guid):
        remove_backup()

    response_jnode = reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
    return response_jnode


def handle_get_profile(profile_jnode_abspath):
    """
    :type profile_jnode_abspath: str
    :rtype: dict
    """
    active_profile_guid = get_active_profile_guid()

    with open(profile_jnode_abspath, 'r') as json_file:
        profile_jnode = ujson.load(json_file)

    profile_jnode['active'] = profile_jnode['id'] == active_profile_guid
    profile_jnode['ca-cert'] = ""
    profile_jnode['client-cert'] = ""
    profile_jnode['client-key'] = ""

    return reply_done(profile_jnode)


def handle_get_profile_set(profile_list):
    """
    :type profile_list: list[str]
    :rtype: dict
    """
    active_profile_guid = get_active_profile_guid()

    profile_set_jnode = []
    for profile_jnode_relpath in profile_list:
        profile_jnode_abspath = SETUPSTORE_VPN_PROFILE_DIRPATH + profile_jnode_relpath
        with open(profile_jnode_abspath, 'r') as profile_jnode_file:
            profile_jnode = ujson.load(profile_jnode_file)

        profile_active = profile_jnode['id'] == active_profile_guid
        profile_set_jnode.append({
            'id': profile_jnode['id'],
            'internal': profile_jnode['internal'],
            'name': profile_jnode['name'], 'active': profile_active,
            'server-host': profile_jnode['server-host'],
            'server-port': profile_jnode['server-port'],
            'protocol': profile_jnode['protocol']
        })

    return reply_done(profile_set_jnode)


def handle_post_profile_active(profile_guid):
    if not try_lock_pid_file(LOCK_FILEPATH):
        return reply_fail(REPLY_CONFLICT_409)

    set_autostart_profile(profile_guid)

    _openvpn_reload_restart()

    return reply_done([])


def handle_post_profile(new_jnode):
    if not try_lock_pid_file(LOCK_FILEPATH):
        return reply_fail(REPLY_CONFLICT_409)

    profile_guid = str(uuid.uuid4())
    # noinspection PyBroadException
    try:
        new_jnode['id'] = profile_guid

        edit_vpn_profile_conf(profile_guid, new_jnode)
        if new_jnode['active']:
            set_autostart_profile(profile_guid)

        with open(SETUPSTORE_VPN_PROFILE_DIRPATH + profile_guid + '.json', 'w') as f:
            f.write((ujson.dumps(new_jnode, indent=2)))

        _openvpn_reload_restart()

    except Exception:
        response_jnode = reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
        return response_jnode

    return reply_done(new_jnode)


def handle_delete_profile(profile_guid):
    if not try_lock_pid_file(LOCK_FILEPATH):
        return reply_fail(REPLY_CONFLICT_409)

    active_profile_guid = get_active_profile_guid()
    if active_profile_guid == profile_guid:
        return reply_fail(REPLY_METHOD_NOT_ALLOWED_405)

    profile_jnode = get_profile_jnode(profile_guid)
    if profile_jnode and profile_jnode['internal']:
        return reply_fail(REPLY_BAD_REQUEST_400)

    # noinspection PyBroadException
    try:
        profile_files = [ETC_OPENVPN_DIRPATH + profile_guid + '.cert',
                         ETC_OPENVPN_DIRPATH + profile_guid + '.ca',
                         ETC_OPENVPN_DIRPATH + profile_guid + '.key',
                         ETC_OPENVPN_DIRPATH + profile_guid + '.conf',
                         SETUPSTORE_VPN_PROFILE_DIRPATH + profile_guid + '.json']

        for profile_file in profile_files:
            if os.path.exists(profile_file):
                os.remove(profile_file)

        _openvpn_reload_restart()

        return reply_done([])

    except Exception:
        pass

    response_jnode = reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
    return response_jnode


def handle_delete_profile_active(profile_guid):
    if not try_lock_pid_file(LOCK_FILEPATH):
        return reply_fail(REPLY_CONFLICT_409)

    active_profile_guid = get_active_profile_guid()
    if profile_guid == active_profile_guid:
        set_autostart_profile(None)

        _openvpn_reload_restart()

    return reply_done([])


def _openvpn_reload_restart():
    os.system('service openvpn stop')
    os.system('systemctl daemon-reload')
    os.system('service openvpn start')


class PathInfo:
    _HTTP_GET_MASK = 0 << 2
    _HTTP_PUT_MASK = 1 << 2
    _HTTP_POST_MASK = 2 << 2
    _HTTP_DELETE_MASK = 3 << 2

    _PROFILE_GUID_MASK = 1 << 1
    _PROFILE_ACTIVE_MASK = 1 << 0

    MODE_UNKNOWN_METHOD = 0x80
    MODE_UNKNOWN_URL = 0x81
    MODE_GET_PROFILE = _HTTP_GET_MASK
    MODE_POST_PROFILE = _HTTP_POST_MASK
    MODE_GET_PROFILE_GUID = _HTTP_GET_MASK | _PROFILE_GUID_MASK
    MODE_PUT_PROFILE_GUID = _HTTP_PUT_MASK | _PROFILE_GUID_MASK
    MODE_DELETE_PROFILE_GUID = _HTTP_DELETE_MASK | _PROFILE_GUID_MASK
    MODE_POST_PROFILE_GUID_ACTIVE = _HTTP_POST_MASK | _PROFILE_GUID_MASK | _PROFILE_ACTIVE_MASK
    MODE_DELETE_PROFILE_GUID_ACTIVE = _HTTP_DELETE_MASK | _PROFILE_GUID_MASK | _PROFILE_ACTIVE_MASK

    def __init__(self, url_mode, profile_guid, profile_active):
        """
        :type url_mode: int
        :type profile_guid: str | None
        :type profile_active: bool | None
        """
        self.url_mode = url_mode
        self.profile_guid = profile_guid
        self.profile_active = profile_active

    @staticmethod
    def deserialize(raw_http_method, raw_path_info):
        """
        :type raw_http_method: str
        :type raw_path_info: str
        :rtype: PathInfo
        """
        url_mode = 0
        profile_guid = None
        profile_active = None

        unknown_url = False
        if raw_path_info:
            path_info_set = raw_path_info.split('/')
            path_info_len = len(path_info_set)

            if path_info_len > 1:
                profile_guid = path_info_set[1]

                try:
                    uuid.UUID('{' + profile_guid + '}')
                    url_mode += PathInfo._PROFILE_GUID_MASK
                except ValueError:
                    unknown_url = True

            if path_info_len > 2 and 'active' == path_info_set[2]:
                url_mode += PathInfo._PROFILE_ACTIVE_MASK

            if path_info_len > 3:
                unknown_url = True

        if unknown_url:
            return PathInfo(PathInfo.MODE_UNKNOWN_URL, None, None)

        mode_list = [PathInfo.MODE_GET_PROFILE,
                     PathInfo.MODE_POST_PROFILE,
                     PathInfo.MODE_GET_PROFILE_GUID,
                     PathInfo.MODE_PUT_PROFILE_GUID,
                     PathInfo.MODE_DELETE_PROFILE_GUID,
                     PathInfo.MODE_POST_PROFILE_GUID_ACTIVE,
                     PathInfo.MODE_DELETE_PROFILE_GUID_ACTIVE]

        if raw_http_method == 'GET' and (PathInfo._HTTP_GET_MASK + url_mode) in mode_list:
            url_mode += PathInfo._HTTP_GET_MASK
        elif raw_http_method == 'PUT' and (PathInfo._HTTP_PUT_MASK + url_mode) in mode_list:
            url_mode += PathInfo._HTTP_PUT_MASK
        elif raw_http_method == 'POST' and (PathInfo._HTTP_POST_MASK + url_mode) in mode_list:
            url_mode += PathInfo._HTTP_POST_MASK
        elif raw_http_method == 'DELETE' and (PathInfo._HTTP_DELETE_MASK + url_mode) in mode_list:
            url_mode += PathInfo._HTTP_DELETE_MASK
        else:
            return PathInfo(PathInfo.MODE_UNKNOWN_METHOD, None, None)

        path_info = PathInfo(url_mode, profile_guid, profile_active)
        return path_info


def _run(http_method, path_info):
    # make schema symlink
    if not os.path.islink(UI_VPN_SCHEMA_FILEPATH) or not os.path.exists(os.readlink(UI_VPN_SCHEMA_FILEPATH)):
        ln_schema2(UI_VPN_SCHEMA_FILEPATH, UI_VPN_DEFAULT_SCHEMA_FILEPATH)

    try:
        with open(UI_VPN_SCHEMA_FILEPATH, 'r') as schema_file:
            request_schema_jnode = ujson.load(schema_file)

    except (IOError, ValueError):
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    path_info = PathInfo.deserialize(http_method, path_info)
    if path_info.url_mode == PathInfo.MODE_UNKNOWN_METHOD:
        return reply_fail(REPLY_METHOD_NOT_ALLOWED_405)

    elif path_info.url_mode == PathInfo.MODE_UNKNOWN_URL:
        return reply_fail(REPLY_NOT_FOUND_404)

    # Dispatch http-method.
    elif path_info.url_mode == PathInfo.MODE_GET_PROFILE:
        if not os.path.exists(SETUPSTORE_VPN_PROFILE_DIRPATH):
            silent_mkdir_r(SETUPSTORE_VPN_PROFILE_DIRPATH)

        profile_list = os.listdir(SETUPSTORE_VPN_PROFILE_DIRPATH)

        response_jnode = handle_get_profile_set(profile_list)

    elif path_info.url_mode == PathInfo.MODE_GET_PROFILE_GUID:
        profile_jnode_abspath = SETUPSTORE_VPN_PROFILE_DIRPATH + path_info.profile_guid + '.json'
        if not os.path.exists(profile_jnode_abspath):
            eventlog.log_d('fail: vpn-profile (%s) not found' % profile_jnode_abspath)
            return reply_fail(REPLY_NOT_FOUND_404)

        response_jnode = handle_get_profile(profile_jnode_abspath)

    elif path_info.url_mode == PathInfo.MODE_PUT_PROFILE_GUID:
        try:
            profile_jnode_abspath = SETUPSTORE_VPN_PROFILE_DIRPATH + path_info.profile_guid + '.json'

            if not os.path.exists(profile_jnode_abspath):
                eventlog.log_d('fail: vpn-profile (%s) not found' % profile_jnode_abspath)
                return reply_fail(REPLY_NOT_FOUND_404)

            with open(profile_jnode_abspath, 'r') as settings_file:
                old_jnode = ujson.load(settings_file)

            new_jnode = ujson.loads(sys.stdin.read())
            jsonschema.validate(new_jnode, request_schema_jnode)

            response_jnode = handle_put_profile(path_info.profile_guid, old_jnode, new_jnode)

        except (ValueError, jsonschema.ValidationError):
            response_jnode = reply_fail(REPLY_BAD_REQUEST_400)

    elif path_info.url_mode == PathInfo.MODE_POST_PROFILE:
        try:
            new_jnode = ujson.loads(sys.stdin.read())
            jsonschema.validate(new_jnode, request_schema_jnode)
            response_jnode = handle_post_profile(new_jnode)

        except (ValueError, jsonschema.ValidationError):
            response_jnode = reply_fail(REPLY_BAD_REQUEST_400)

    elif path_info.url_mode == PathInfo.MODE_POST_PROFILE_GUID_ACTIVE:
        response_jnode = handle_post_profile_active(path_info.profile_guid)

    elif path_info.url_mode == PathInfo.MODE_DELETE_PROFILE_GUID:
        profile_jnode_abspath = SETUPSTORE_VPN_PROFILE_DIRPATH + path_info.profile_guid + '.json'
        if not os.path.exists(profile_jnode_abspath):
            return reply_fail(REPLY_NOT_FOUND_404)

        response_jnode = handle_delete_profile(path_info.profile_guid)

    elif path_info.url_mode == PathInfo.MODE_DELETE_PROFILE_GUID_ACTIVE:
        response_jnode = handle_delete_profile_active(path_info.profile_guid)

    else:
        response_jnode = reply_fail(REPLY_METHOD_NOT_ALLOWED_405)

    return response_jnode


def _main():
    eventlog.open_log('cgi-vpn', 'debug')
    exit_code = 1

    # noinspection PyBroadException
    try:
        env_http_mode = os.environ['REQUEST_METHOD']
        env_path_info = os.environ['PATH_INFO']

        response_dict = _run(env_http_mode, env_path_info)
        print ujson.dumps(dict(response_dict))
        exit_code = 0

    except Exception:
        eventlog.log_x()
        print ujson.dumps(reply_fail(REPLY_INTERNAL_SERVER_ERROR_500))

    finally:
        eventlog.close_log()

    sys.exit(exit_code)


if __name__ == '__main__':
    _main()
