import base64
import datetime
import ujson
import urllib2

import pytz
from twisted.internet import reactor
from twisted.internet.defer import DeferredList, inlineCallbacks, returnValue
from twisted.web.client import Agent, readBody
from twisted.web.http_headers import Headers

from edge import eventlog
from edge.image.image import ImEnv, ImEngine
from edge.image.types import ImFooterVars, MemBlob
from edge.service.tz import isostr_to_datetime
from edge_bbox.types import mk_plate_bbox, RectF

from .types import SectionSpeedEvidence, SectionArgs, DetectorArgs, RuleArgs
from ..com.schedule import Schedule
from ..types import IDetector
from ...model.models import Track, TrTrackItem
from ...model.types import BestWaypoint, Event, Waypoint
from ...services.vs import VsImageInfo, VsImageBufferError


FEED_QUERY = (u'{url}/api/v2/origins/event/feed?origin={origin}&type=transit&'
              'dir={dir}&position=required&plate-method=exact&'
              'plate-pattern={plate}&limit=1')
DIST_QUERY = ('{url}/api/v2/origins/event/dist/{event_id}'
              '/by-ref/{img_name}')


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


@inlineCallbacks
def _get_file(agent, event_id, img_name, base_url, user, password):
    """
    :type event_id: str
    :type img_name: str
    :type base_url: str
    :type user: str
    :type password: str
    :rtype: bytes
    """
    url = DIST_QUERY.format(url=base_url, event_id=event_id, img_name=img_name)
    base64string = base64.b64encode('%s:%s' % (user, password))
    resp = yield agent.request(b'GET', url,
            Headers({"Authorization": ["Basic %s" % base64string]}), None)
    if resp.code != 200:
        eventlog.log_e('fail: get file: <{0}>, <{1}>'.format(url, resp.code))
        returnValue(None)
    file_contents = yield readBody(resp)
    returnValue(file_contents)


@inlineCallbacks
def _get_feed_record(agent, base_url, origin, direction, plate_text, user, password):
    """
    :type base_url: str
    :type origin: str
    :type direction: Dir
    :type plate_text: str
    :type user: str
    :type password: str
    :rtype: dict | None
    """
    url = FEED_QUERY.format(url=base_url, origin=origin, dir=direction,
                            plate=plate_text)
    url = urllib2.quote(url.encode('utf8'), ':/&?=')
    base64string = base64.b64encode('%s:%s' % (user, password))
    resp = yield agent.request(b'GET', url,
                               Headers({"Authorization": [
                                   "Basic %s" % base64string]}), None)
    if resp.code != 200:
        eventlog.log_e('fail: get feed record: <{0}>, <{1}>'.format(url, resp.code))
        returnValue(None)

    feed_str = yield readBody(resp)
    feed_record = ujson.loads(feed_str)  # list of feed records
    returnValue(feed_record)


def _time_jnode_to_timestamp(time_jnode):
    """
    :type time_jnode: str
    :rtype: float
    """
    dt = isostr_to_datetime(time_jnode)
    ts = (dt - pytz.utc.localize(datetime.datetime(1970, 1, 1)).astimezone(pytz.utc)).total_seconds()
    return ts  # timestamp in seconds


def _get_section_speed(area_d, feed_reply, track, best_waypoint):
    """
    :param area_d: distance in meters
    :type area_d: int
    :type feed_reply: dict
    :type track: Track
    :type best_waypoint: BestWaypoint
    :rtype: float
    """
    init_t_str = feed_reply[0]['event']['time']
    init_t = _time_jnode_to_timestamp(init_t_str)  # s
    fini_t = best_waypoint.ts  # ms

    init_d = feed_reply[0]['event']['vehicle']['position']['forward']  # cm
    if area_d >= 1000:
        if track.ra_track:
            fini_d = track.ra_track[0].raw_f
        else:
            fini_d = 0
    else:
        fini_d = track.ra_track[0].raw_f  # cm
    dist_km = (area_d - init_d / 100.0 + fini_d / 100) / 1000
    time_h = (fini_t / 1000000.0 - init_t) / 3600
    section_speed = dist_km / time_h
    return section_speed


class _RuleServices:
    def __init__(self, tz_service, vs_service, about_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        """
        self.tz = tz_service
        self.vs = vs_service
        self.about = about_service


class _SceneMsRef(VsImageInfo):
    def __init__(self, tr_item):
        """
        :type tr_item: TrTrackItem
        """
        VsImageInfo.__init__(self, tr_item.ts)
        self.tr_item = tr_item


class Rule:
    def __init__(self, services, rule_args, detector_name, section_args,
                 fini_im_env):
        """
        :type services: _RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :type section_args: SectionArgs
        :type fini_im_env: ImEnv
        """
        self._detector_name = detector_name

        self._services = services
        self._scheduler = Schedule(rule_args.schedule, services.tz)
        self._args = rule_args

        self._section_args = section_args
        self._fini_im_env = fini_im_env

        self.agent = Agent(reactor, connectTimeout=60)

    @inlineCallbacks
    def check(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        :rtype: SectionSpeedEvidence | None
        """
        cts = track.lr_track.min_ts
        lts = track.lr_track.max_ts
        if not(self._scheduler.check2(lts) or self._scheduler.check2(cts)):
            returnValue(None)

        url = self._section_args.init_area.controller
        origin = self._section_args.init_area.origin
        direction = self._section_args.direction
        user, password = self._section_args.init_area.auth
        plate_text = base.get_plate_text()
        try:
            feed_record = yield _get_feed_record(self.agent,
                                                 url, origin, direction,
                                                 plate_text, user, password)
        except Exception:
            eventlog.log_e('fail: get remote feed record')
            returnValue(None)
        if not feed_record:  # !200 or empty list
            returnValue(None)

        measurements_fini = base.measurements
        waypoints = measurements_fini[0].waypoints
        best_waypoint_fini = [w for w in waypoints
                              if isinstance(w, BestWaypoint)][0]

        area_d = self._section_args.dist  # m
        if area_d < 1000 and not track.ra_track:
            eventlog.log_i('Distance must be > 1000 to measure without radar')
            returnValue(None)

        try:
            section_speed = _get_section_speed(area_d, feed_record, track,
                                               best_waypoint_fini)
        except KeyError:
            returnValue(None)

        speed_limit = self._args.limits.speed_limit
        max_overspeed = self._args.limits.over_speed
        if not (section_speed > speed_limit + max_overspeed):
            returnValue(None)
        section_speed = round(section_speed, 1)

        # download arg-json with measurements; download images
        try:
            event_id = feed_record[0]['event']['uuid']
            args_json_name = feed_record[0]["dist"]["args"]
        except KeyError:
            returnValue(None)

        try:
            args_bytes = yield _get_file(self.agent,
                                         event_id, args_json_name, url, user,
                                         password)
            args_json = ujson.loads(args_bytes)
            measurements = args_json['measurements'][0]
            images = measurements['waypoints']['best']['images']
            scene_img_name = images['scene']['name']
            plate_img_name = images['license-plates'][0]['name']

            scene_img = yield _get_file(self.agent,
                                        event_id, scene_img_name, url, user,
                                        password)
            plate_img = yield _get_file(self.agent,
                                        event_id, plate_img_name, url, user,
                                        password)
        except Exception:
            returnValue(None)

        scene_ms_refs = map(_SceneMsRef, track.tr_track)
        fini_scene_ms_refs = scene_ms_refs[:-6:-1]
        try:
            fini_winner_img = self._services.vs.export_image_rgba(self._fini_im_env.scene_rgba, fini_scene_ms_refs)
        except VsImageBufferError as ex:
            self._fini_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._fini_im_env.footer_h)
            fini_winner_img = self._services.vs.export_image_rgba(self._fini_im_env.scene_rgba, fini_scene_ms_refs)

        if not fini_winner_img:
            eventlog.log_d('fail: get init/fini image from vs')
            returnValue(None)

        best_waypoint_init_jnode = measurements['waypoints']['best']
        evidence = self._mk_evidence(section_speed, best_waypoint_init_jnode,
                                     scene_img, plate_img, plate_text,
                                     speed_limit, track,
                                     fini_winner_img.tr_item)
        returnValue(evidence)

    def _mk_evidence(self, section_speed, init_wp_jnode,  init_scene,
                     init_plate, plate_text, speed_limit, track,
                     fini_tr_item):
        """
        :type section_speed: float
        :type init_wp_jnode: dict
        :type init_scene: bytes
        :type init_plate: bytes
        :type plate_text: str
        :type speed_limit: int
        :type track: Track
        :type fini_tr_item: TrTrackItem

        :rtype: SectionSpeedEvidence | None
        """
        def mk_waypoint_feed_record():
            """
            :rtype: Waypoint
            """
            init_ts = _time_jnode_to_timestamp(init_wp_jnode['time'])
            init_ts = long(init_ts * 1000000)
            wp = Waypoint('init', init_ts)

            scene = MemBlob('scene', init_ts, bytes(init_scene))
            plate = MemBlob('plate', init_ts, bytes(init_plate))

            wp.add_images([scene, plate])
            return wp

        init_wp = mk_waypoint_feed_record()

        direction = self._section_args.direction

        plate_rect = RectF(*fini_tr_item.lr_item.plate_quad.get_rel_bbox(
            track.lr_track.best_image.width,
            track.lr_track.best_image.height))
        plate_bbox = mk_plate_bbox(plate_rect,
                                   *self._fini_im_env.scene_rgba.wh)

        waypoint_ts = self._fini_im_env.scene_rgba.ts
        waypoint_origin = self._services.about.get_origin(waypoint_ts)
        footer_vars = ImFooterVars(self._services.tz, self._services.about,
                                   waypoint_ts,
                                   self._fini_im_env.footer_def_vars,
                                   location=waypoint_origin.location,
                                   geolocation=waypoint_origin.geolocation,
                                   plate_text=plate_text,
                                   section_speed=section_speed,
                                   direction=direction, max_speed=speed_limit)

        im_blobs = self._fini_im_env.engine.mk_blobs(
            self._fini_im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track,
            fini_tr_item.lr_item.plate_quad, scene=True, plate=True)

        if len(im_blobs) != 2:
            eventlog.log_d('fail: get all blobs for evidence')
            raise RuntimeError

        fini_wp = Waypoint('fini', waypoint_ts)
        fini_wp.add_images(im_blobs)

        evidence = SectionSpeedEvidence(
            self._services.about.get_origin(waypoint_ts), waypoint_ts,
            section_speed, self._args, self._detector_name, direction)
        evidence.add_waypoints([
            init_wp,
            fini_wp
        ])

        return evidence


class _Detector(IDetector):
    def __init__(self, tz_service, vs_service, about_service, args):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type args: DetectorArgs
        """
        self._identity = args.identity

        self._handler = None

        self._fini_im_env = ImEnv(
            ImEngine(args.im_engine),
            args.im_engine.scene.footer.height,
            args.def_footer_mapping)

        self._rules = []
        rule_services = _RuleServices(tz_service, vs_service, about_service)
        for rule in args.rules:
            if rule.enabled:
                self._rules.append(Rule(rule_services, rule,
                                        args.name, args.section,
                                        self._fini_im_env))
        self._section = args.section

        IDetector.__init__(self)

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """
        def log_exc(e):
            eventlog.log_w(str(e))

        def handle_results(d_results):
            """
            :param d_results: list of pairs (deferred success, deferred result)
            :type d_results: list (bool, SectionSpeedEvidence | None)
            """
            for success, evidence in d_results:
                if success and evidence:
                    self._handler(Event(self._identity, base, evidence))

        if track.lr_track.direction != self._section.direction.dir:
            return

        deferreds = []
        for rule in self._rules:
            deferreds.append(rule.check(track, base))

        d_list = DeferredList(deferreds)
        d_list.addCallbacks(handle_results, log_exc)

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler
